"""Standalone functions for benchmarking editor performance across metrics."""
import logging
import random
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import Any, Callable, Sequence, cast

from evaluation import data, metrics, models, precompute
from evaluation.evaluator import * 
from evaluation.utils import tokenizer_utils
from evaluation.utils.typing import Dataset, Device, StrSequence
from pastalib import pasta

import numpy as np
import scipy.stats
import torch
import torch.utils.data
from dataclasses_json import DataClassJsonMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import accuracy_score, f1_score, matthews_corrcoef
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)

DEFAULT_PROMPT_PREFIX = "The following is an excerpt from a Wikipedia article:\n\n"
DEFAULT_PROMPT_TEMPLATE = "{} is"
DEFAULT_MAX_LENGTH = 100
DEFAULT_MAX_LENGTH_ERROR_CORRECTION = 150
DEFAULT_TOP_K = 3
DEFAULT_N_TOP_TOKENS = DEFAULT_TOP_K
DEFAULT_BATCH_SIZE = 16


@dataclass(frozen=True)
class EfficacySample(DataClassJsonMixin):
    """Wrapper around a single efficacy sample."""

    id: str
    prompt: str
    target_score: float
    comparator_score: float


@dataclass(frozen=True)
class EfficacyBenchmarkResults(DataClassJsonMixin):
    """Wrapper around efficacy benchmark results."""

    samples: list[EfficacySample]
    metrics: metrics.EfficacyMetrics


@dataclass(frozen=True)
class CounterFactEvaluationResult(DataClassJsonMixin):
    """Result for single sample from `Editor.evaluate`."""

    sample: dict

    top_tokens: list[str] | None = None
    top_logps: list[float] | None = None
    generations: list[str] | None = None

    target_mediated_score: float | None = None
    target_unmediated_score: float | None = None


@dataclass(frozen=True)
class CounterFactEvaluateRun(DataClassJsonMixin):
    """Wrapper around a list of individual evaluation results."""

    results: list[CounterFactEvaluationResult]


@torch.inference_mode()
def counterfact_evaluate(
    mt, 
    dataset: Dataset,
    pasta_steerer: pasta.PASTA|None = None, 
    batch_size: int = 16,
    n_top: int = 3,
    max_length: int | None = None,
    max_new_tokens: int | None = None,
    desc: str|None = None,
    device: Device|None = None,
    return_mediated: bool = True,
    return_unmediated: bool = True,
    add_unmediated_fact: bool = True,
    add_few_shot: int | None = None, 
    few_shot_index: str | None = None,  
    add_marker: str | None = None, 
) -> CounterFactEvaluateRun:
    """
    Evaluation on CounterFact 

    Args:
        mt: The model to be evaluated. 
        dataset: Dataset to evaluate on.
        pasta_steerer: The PASTA steerer for attention steering. 
        batch_size: Model batch size.
        n_top: Number of top words/probs to return.
        max_length: Number of tokens to generate including prompt.
        max_new_tokens: Number of tokens to generate not including prompt.
        desc: The tqdm description.
        device: Send all data to this device. Defaults to None.
        return_mediated: Return mediated token probability.
        return_unmediated: Return unmediated token probability.
        add_unmediated_fact: Whether to present the model both new and old facts.  

    Returns:
        The evaluation results, one per entry in the dataset.

    """
    mt.eval_()
    mt.to_(device)
    include_target_probs = "target_mediated" in dataset.column_names
    if desc is None:
        desc = f"Evaluate CounterFact"
    if max_length is None and max_new_tokens is None:
        max_length = DEFAULT_MAX_LENGTH

    exclude_columns = []
    if not return_mediated:
        exclude_columns.append("target_mediated")
    if not return_unmediated:
        exclude_columns.append("target_unmediated")
    columns = data.column_names(dataset, exclude=exclude_columns)

    if add_few_shot is not None and add_few_shot > 0: 
        if few_shot_index is not None:
            few_shot_index = [int(idx) for idx in few_shot_index.split(",")]
            assert len(few_shot_index) == add_few_shot
        else:
            few_shot_index = np.random.randint(len(dataset), size=add_few_shot).tolist()
        fewshot_examples = precompute.prepare_counterfact_few_shot_examples(
            dataset.select(few_shot_index), add_unmediated_fact=add_unmediated_fact, 
        )
    else:
        fewshot_examples = None 

    results = []
    with dataset.formatted_as("torch", columns=columns):
        loader = torch.utils.data.DataLoader(
            cast(torch.utils.data.Dataset, dataset), batch_size=batch_size
        )
        for batch in tqdm(loader, desc=desc):
            if not precompute.has_editor_inputs(batch):
                batch.update(
                    precompute.editor_inputs_from_batch(
                        mt=mt, batch=batch, device=device, 
                        return_target_token_ids=return_mediated or return_unmediated,
                        target_token_first_space=True if "gpt" in mt.model.name_or_path else False,
                    )
                )

            prompts = batch["prompt"]
            prompts = batch["prompt"]
            contexts = batch["context"]
            attributes = batch["attribute"]
            targets_mediated = batch["target_mediated"]
            targets_unmediated = batch["target_unmediated"]
            current_batch_size = len(prompts)

            if add_unmediated_fact:
                # Modify the prompts and provide the model both new and old facts 
                new_prompts = []
                for prompt, context, target_mediated, target_unmediated in zip(
                    prompts, contexts, targets_mediated, targets_unmediated
                ):
                    unmediated_prefix = "Previously "
                    mediated_prefix = "Currently "
                    unmediated_fact = context.replace(target_mediated, target_unmediated)+". "
                    new_prompt = f"{unmediated_prefix}{unmediated_fact}{mediated_prefix}{prompt}"

                    new_prompts.append(new_prompt)
                prompts = new_prompts

            if add_marker is not None:
                prompts = [
                    prompt.replace(attr, add_marker+attr+add_marker) for prompt,attr in zip(prompts, attributes)
                ]
            if fewshot_examples is not None:
                prompts = [
                    fewshot_examples + prompt for prompt in prompts
                ]
            
            with models.set_padding_side(mt, padding_side="left"):
                inputs, prompt_offset_mapping = precompute.inputs_from_batch(mt, prompts, device=device)
            generate_kwargs = dict(
                do_sample=False,
                return_dict_in_generate=True,
                output_scores=True,
                max_length=max_length,
                max_new_tokens=max_new_tokens,
                pad_token_id=mt.tokenizer.eos_token_id,
            )
            # Here, we emphasize the new facts (attributes) with PASTA
            if pasta_steerer is not None:
                with pasta_steerer.apply_steering(
                    model=mt.model, 
                    strings=prompts, 
                    substrings=attributes, 
                    model_input=inputs, 
                    offsets_mapping=prompt_offset_mapping
                ) as steered_model: 
                    outputs = steered_model.generate(**inputs, **generate_kwargs)
            else:
                outputs = mt.model.generate(**inputs, **generate_kwargs)
            
            batched_results: dict = {}
            first_token_logps = torch.log_softmax(outputs.scores[0], dim=-1)
            top_logps, top_token_ids = first_token_logps.topk(k=n_top, dim=-1)
            top_tokens = tokenizer_utils.batch_convert_ids_to_tokens(top_token_ids, mt.tokenizer)
            generations = mt.tokenizer.batch_decode(outputs.sequences, skip_special_tokens=True)

            batched_results[f"top_logps"] = top_logps.tolist()
            batched_results[f"top_tokens"] = top_tokens
            batched_results[f"generations"] = [[g] for g in generations]
            if include_target_probs:
                target_keys = []
                if return_mediated:
                    target_keys.append("mediated")
                if return_unmediated:
                    target_keys.append("unmediated")
                batch_indices = torch.arange(current_batch_size)
                for target_key in target_keys:
                    target_id = batch[f"target_{target_key}.token_id"]
                    target_probs = first_token_logps[batch_indices, target_id]
                    target_prob_key = f"target_{target_key}_score"
                    batched_results[target_prob_key] = target_probs.tolist()

            for bi in range(current_batch_size):
                result: dict = {k: vs[bi] for k, vs in batched_results.items()}
                results.append(result)

    # Finally, decorate results with original sample data.
    assert len(results) == len(dataset)
    for sample, result in zip(dataset, results):
        result.update(
            sample={
                key: sample[key]
                for key in data.ContextMediationSample.__required_keys__
            }
        )

    return CounterFactEvaluateRun([CounterFactEvaluationResult(**r) for r in results])



@torch.inference_mode()
def counterfact_efficacy(
    *,
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    pasta_steerer: pasta.PASTA|None = None, 
    desc: str | None = None,
    **kwargs: Any,
) -> EfficacyBenchmarkResults:
    """Run the CounterFact efficacy benchmark.

    Args:
        mt: The model to evaluate. 
        dataset: The dataset to evaluate on.
        pasta_steerer: The PASTA steerer to steer the model generaion.

    Returns:
        Evaluation results 
    
    """
    if desc is None:
        desc = "efficacy benchmark"

    evaluate_kwargs = dict(max_new_tokens=1, desc=desc, **kwargs)
    run = counterfact_evaluate(mt, dataset, pasta_steerer=pasta_steerer, **evaluate_kwargs)
    target_score_key = "target_mediated_score"
    comparator_score_key = "target_unmediated_score"

    samples = []
    for result in run.results:
        sid = result.sample["id"]
        prompt = result.sample["prompt"]

        target_score = getattr(result, target_score_key)
        assert target_score is not None

        comparator_score = getattr(result, comparator_score_key)
        assert comparator_score is not None

        logger.debug(f"ID={sid} SCORE_T={target_score} SCORE_COMP={comparator_score}")
        sample = EfficacySample(
            id=sid,
            prompt=prompt,
            target_score=target_score,
            comparator_score=comparator_score,
        )
        samples.append(sample)

    efficacy_metrics = metrics.efficacy(
        [[sample.target_score] for sample in samples],
        [[sample.comparator_score] for sample in samples],
        store_values=False,
    )
    return EfficacyBenchmarkResults(samples=samples, metrics=efficacy_metrics)


@dataclass(frozen=True)
class ParaphraseSample(DataClassJsonMixin):
    """Wrapper around a single paraphrase benchmark sample."""

    id: str
    prompts: list[EfficacySample]
    efficacy_score: float
    efficacy_magnitude: float


@dataclass(frozen=True)
class CounterFactParaphraseBenchmarkResults(DataClassJsonMixin):
    """Wrapper around paraphrase benchmark results."""

    samples: list[ParaphraseSample]
    metrics: metrics.EfficacyMetrics


@torch.inference_mode()
def counterfact_paraphrase(
    *,
    mt: models.ModelAndTokenizer | None = None,
    pasta_steerer: pasta.PASTA|None = None,
    dataset: Dataset,
    desc: str | None = None,
    **kwargs: Any,
) -> CounterFactParaphraseBenchmarkResults:
    """Run the CounterFact paraphrase benchmark.

    Since this benchmark relies on extra data, it can only be used with the CounterFact
    dataset. The `counterfact_generation` benchmark is like this as well.

    This function expects that each sample in the dataset supports an access like:

        prompts = sample["source"]["generation_prompts"]

    """
    if desc is None:
        desc = "paraphrase benchmark"
    dataset = _counterfact_select_and_flatten(
        dataset, "paraphrase_prompts", desc=f"{desc} [flatten dataset]"
    )
    efficacy_benchmark = counterfact_efficacy(
        mt=mt,
        pasta_steerer=pasta_steerer,
        dataset=dataset,
        desc=desc,
        **kwargs,
    )

    results_by_sample_id: dict = defaultdict(list)
    for result in efficacy_benchmark.samples:
        results_by_sample_id[result.id].append(result)
    results_by_sample_id = OrderedDict(results_by_sample_id)

    efficacy_metrics = metrics.efficacy(
        [
            [result.target_score for result in results]
            for results in results_by_sample_id.values()
        ],
        [
            [result.comparator_score for result in results]
            for results in results_by_sample_id.values()
        ],
    )

    # Reformat EfficacySample -> ParaphraseSample
    samples = []
    for (sid, results), efficacy_score, efficacy_magnitude in zip(
        results_by_sample_id.items(),
        cast(list, efficacy_metrics.score.values),
        cast(list, efficacy_metrics.magnitude.values),
    ):
        sample = ParaphraseSample(
            id=sid,
            prompts=results,
            efficacy_score=efficacy_score,
            efficacy_magnitude=efficacy_magnitude,
        )
        samples.append(sample)

    return CounterFactParaphraseBenchmarkResults(
        samples=samples,
        metrics=efficacy_metrics.without_values(),
    )


@dataclass(frozen=True)
class GenerationSample(DataClassJsonMixin):
    """Wrapper around a single sample from the generation benchmark."""

    id: str
    generations: list[str]
    references: list[str]
    fluency_score: float
    consistency_score: float


@dataclass(frozen=True)
class GenerationMetrics(DataClassJsonMixin):
    """Wrapper around all generation metrics."""

    fluency: metrics.Metric
    consistency: metrics.Metric


@dataclass(frozen=True)
class CounterFactGenerationBenchmarkResults(DataClassJsonMixin):
    """Wrapper around generation benchmark results."""

    samples: list[GenerationSample]
    metrics: GenerationMetrics


@torch.inference_mode()
def counterfact_generation(
    *,
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    pasta_steerer: pasta.PASTA|None = None,
    attribute_snippets: data.AttributeSnippets | None = None,
    tfidf_vectorizer: TfidfVectorizer | None = None,
    max_length: int | None = None,
    max_new_tokens: int | None = None,
    desc: str | None = None,
    **kwargs: Any,
) -> CounterFactGenerationBenchmarkResults:
    """Run the CounterFact generation benchmark.

    Free-form generates on several "generation prompts" per sample, and records
    the fluency of the generations (measured by weighted n-gram entropy) and
    consistency with other texts about entities with the same attribute.

    This benchmark *requires* the dataset to be CounterFact or something that looks
    like it, since it uses extra data that is specific to CounterFact.

    Specifically, it expects each sample can be accessed like:

        prompts = sample["source"]["generation_prompts"]

    """
    if attribute_snippets is None:
        attribute_snippets = data.load_attribute_snippets()
    if tfidf_vectorizer is None:
        tfidf_vectorizer = data.load_counterfact_tfidf_vectorizer()
    if max_new_tokens is None and max_length is None:
        max_length = DEFAULT_MAX_LENGTH
    if desc is None:
        desc = "generate benchmark"

    dataset = _counterfact_select_and_flatten(
        dataset, "generation_prompts", desc=f"{desc} [flatten dataset]"
    )

    evaluate_kwargs = dict(
        max_new_tokens=max_new_tokens,
        max_length=max_length,
        desc=f"{desc} [run model]",
        **kwargs,
    )
    run = counterfact_evaluate(mt, dataset, pasta_steerer=pasta_steerer, **evaluate_kwargs)
    generations_key = "generations"
    run_results_by_id = _group_results_by_id(run)

    samples = []
    for sid, results in tqdm(run_results_by_id.items(), desc=f"{desc} [tfidf]"):
        result = next(iter(results))
        cf_requested_rewrite = result.sample["source"]["requested_rewrite"]
        relation_id = cf_requested_rewrite["relation_id"]
        target_id = cf_requested_rewrite["target_new"]["id"]

        generations = [getattr(result, generations_key)[0] for result in results]
        references = [
            snippet["text"] for snippet in attribute_snippets[relation_id][target_id]
        ]

        consistency_score = metrics.tfidf_similarity(
            generations, references, tfidf_vectorizer
        )
        fluency_score = metrics.weighted_n_gram_entropy(generations)

        entity = result.sample["entity"]
        attribute = result.sample["attribute"]
        logger.debug(f"ID={sid} ENTITY={entity}, ATTR={attribute}")
        logger.debug(f"ID={sid} REFERENCES={references}")
        logger.debug(f"ID={sid} GENERATIONS={generations}")

        sample = GenerationSample(
            id=sid,
            generations=generations,
            references=references,
            fluency_score=fluency_score,
            consistency_score=consistency_score,
        )
        samples.append(sample)

    fluency = metrics.Metric.aggregate(
        [sample.fluency_score for sample in samples], store_values=False
    )
    consistency = metrics.Metric.aggregate(
        [sample.consistency_score for sample in samples], store_values=False
    )
    generation_metrics = GenerationMetrics(fluency=fluency, consistency=consistency)
    return CounterFactGenerationBenchmarkResults(
        samples=samples, metrics=generation_metrics
    )


def _counterfact_select_and_flatten(
    dataset: Dataset, column: str, desc: str | None = None
) -> Dataset:
    """Select the given column in counterfact, dedupe it, and flatten it."""
    column_names = data.column_names(dataset)

    def select_and_flatten_counterfact_row(row: dict) -> dict:
        prompts = list(set(row["source"][0][column]))
        result = {"prompt": prompts}
        for key in data.ContextMediationSample.__required_keys__:
            if key not in result:
                result[key] = [row[key][0]] * len(prompts)
        return result

    return dataset.map(
        select_and_flatten_counterfact_row,
        batched=True,
        batch_size=1,
        remove_columns=column_names,
        desc=desc,
    )


def _group_results_by_id(results: CounterFactEvaluateRun) -> OrderedDict:
    """Group results by sample ID."""
    grouped = defaultdict(list)
    for result in results.results:
        grouped[result.sample["id"]].append(result)
    return OrderedDict(grouped)


@dataclass(frozen=True)
class BiasBiosEvaluationSample:
    """Wrapper around error correction sample."""

    id: str
    prompt: str
    generation: str

    predictions: list[str]
    target: str

    logp_predictions: list[float]
    logp_target: float

    fluency: float
    consistency: float


@dataclass(frozen=True)
class BiasBiosEvaluationMetrics(DataClassJsonMixin):
    """Wrapper around aggregated error correction metrics."""

    top1_accuracy: float
    topk_accuracy: float
    k: int

    fluency: metrics.Metric
    consistency: metrics.Metric


@dataclass(frozen=True)
class BiasBiosEvaluationResults(DataClassJsonMixin):
    """Wrapper around error correction benchmark."""

    samples: list[BiasBiosEvaluationSample]
    metrics: BiasBiosEvaluationMetrics


@torch.inference_mode()
def biasbios_prediction_evaluation(
    *,
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    pasta_steerer: pasta.PASTA|None = None, 
    tfidf_vectorizer: TfidfVectorizer | None = None,
    references: dict | None = None,
    batch_size: int = DEFAULT_BATCH_SIZE,
    top_k: int = DEFAULT_TOP_K,
    max_length: int | None = None,
    max_new_tokens: int | None = None,
    device: Device | None = None,
    desc: str | None = None,
    add_few_shot: int | None = None, 
    few_shot_index: str | None = None,  
    add_marker: str | None = None, 
    **kwargs,
) -> BiasBiosEvaluationResults:
    """Run BiasBios prediction benchmark.

    This benchmark involves measuring accuracy on the BiasBios prediction task. 

    Args:
        mt: The model to evaluate. 
        dataset: The dataset to evaluate on.
        pasta_steerer: The PASTA steerer to steer the model generaion. 
        tfidf_vectorizer: For computing consistency score.
        references: Mapping from label to reference texts for that label. By default,
            full bios for each label will be used.
        batch_size: Batch size for model.
        top_k: Compute top-k labels predicted by model.
        prompt_key: Which column in dataset to use as prompt.
        max_length: Max sequence length (input+output).
        max_new_tokens: Max number of new tokens to generate. Cannot be used with
            `max_length`, see huggingface docs.
        device: Send model and data to this device.
        desc: TQDM description. 
        add_few_shot: Whether to apply few-shot prompting. 
        few_shot_index: Sample index for few shots. 
        add_marker: Whether to apply marked prompting. 

    Returns:
        Benchmark results.

    """
    if max_length is None and max_new_tokens is None:
        max_length = DEFAULT_MAX_LENGTH_ERROR_CORRECTION
    if desc is None:
        desc = "BiasBios Evaluation"

    if tfidf_vectorizer is None:
        tfidf_vectorizer = data.load_biosbias_tfidf_vectorizer()
    if references is None:
        references = defaultdict(list)
        for sample in dataset:
            references[sample["target_mediated"]].append(sample["source"]["bio"])

    labels = sorted({x["target_mediated"] for x in dataset})
    label_add_space = True if "gpt" in mt.model.name_or_path else False
    labels_token_idx = precompute.first_token_ids_from_batch(mt, labels, add_space=label_add_space)

    reference_tfidfs = {
        key: tfidf_vectorizer.transform(texts).mean(axis=0).A
        for key, texts in tqdm(references.items(), desc=f"{desc} [reference tfidfs]")
    }

    if add_few_shot is not None and add_few_shot > 0: 
        if few_shot_index is not None:
            few_shot_index = [int(idx) for idx in few_shot_index.split(",")]
            assert len(few_shot_index) == add_few_shot
        else:
            few_shot_index = np.random.randint(len(dataset), size=add_few_shot).tolist()
        fewshot_examples = prepare_biasbios_few_shot_examples(
            dataset.select(few_shot_index), text_key="context"
        )

    columns = data.column_names(dataset, exclude=["target_unmediated"])
    with dataset.formatted_as("torch", columns=columns):
        loader = torch.utils.data.DataLoader(
            cast(torch.utils.data.Dataset, dataset),
            batch_size=batch_size,
        )

        samples = []
        for batch in tqdm(loader, desc=desc):
            ids = batch["id"]
            prompts = batch["prompt"]
            targets = batch["target_mediated"]
            targets_idx = precompute.first_token_ids_from_batch(mt, targets, add_space=label_add_space)

            if add_few_shot is not None and add_few_shot > 0:
                batch['prompt'] = [fewshot_examples+prompt for prompt in batch['prompt']]
                prompts = [fewshot_examples+prompt for prompt in prompts]

            if add_marker is not None:
                batch['prompt'] = [
                    prompt.replace(attr, add_marker+attr+add_marker) for prompt,attr in zip(batch['prompt'], batch['attribute'])
                ]
                prompts = batch['prompt']

            with models.set_padding_side(mt, padding_side="left"):
                inputs, prompt_offset_mapping = precompute.inputs_from_batch(mt, prompts, device=device)

            generate_kwargs = dict(
                return_dict_in_generate=True,
                output_scores=True,
                max_length=max_length,
                max_new_tokens=max_new_tokens,
                pad_token_id=mt.tokenizer.eos_token_id,
            )
            if pasta_steerer is not None:
                # Apply PASTA steering
                # Here we emphasize the first sentence `attribute`
                with pasta_steerer.apply_steering(
                    model=mt.model, 
                    strings=prompts, 
                    substrings=batch['attribute'], 
                    model_input=inputs, 
                    offsets_mapping=prompt_offset_mapping
                ) as steered_model: 
                    outputs = steered_model.generate(**inputs, **generate_kwargs)
            else:
                outputs = mt.model.generate(**inputs, **generate_kwargs)

            generations = mt.tokenizer.batch_decode(
                outputs.sequences[:, inputs.input_ids.shape[1] :],
                skip_special_tokens=True,
            )
            distributions = torch.log_softmax(outputs.scores[0], dim=-1)

            for sid, prompt, distribution, generation, target, target_idx in zip(
                ids, prompts, distributions, generations, targets, targets_idx
            ):
                label_log_probs = distribution[labels_token_idx]

                logp_predictions, predictions_idx = label_log_probs.topk(
                    k=top_k, dim=-1
                )
                predictions = [labels[idx] for idx in predictions_idx]

                logp_target = distribution[target_idx]

                fluency_score = metrics.weighted_n_gram_entropy(generation)

                [generation_tfidf] = tfidf_vectorizer.transform([generation]).A
                reference_tfidf = reference_tfidfs[target]
                consistency_score = metrics.vector_similarity(
                    generation_tfidf.squeeze(), reference_tfidf.squeeze()
                )

                sample = BiasBiosEvaluationSample(
                    id=sid,
                    prompt=prompt,
                    generation=generation,
                    predictions=predictions,
                    logp_predictions=logp_predictions.tolist(),
                    target=target,
                    logp_target=logp_target.item(),
                    fluency=fluency_score,
                    consistency=consistency_score,
                )
                samples.append(sample)

    n_correct_top1 = sum(x.predictions[0] == x.target for x in samples)
    n_correct_topk = sum(x.target in x.predictions for x in samples)
    top1_accuracy = n_correct_top1 / len(samples)
    topk_accuracy = n_correct_topk / len(samples)

    fluency = metrics.Metric.aggregate([x.fluency for x in samples], store_values=False)
    consistency = metrics.Metric.aggregate(
        [x.consistency for x in samples], store_values=False
    )

    error_correction_metrics = BiasBiosEvaluationMetrics(
        top1_accuracy=top1_accuracy,
        topk_accuracy=topk_accuracy,
        k=top_k,
        fluency=fluency,
        consistency=consistency,
    )

    return BiasBiosEvaluationResults(
        samples=samples,
        metrics=error_correction_metrics,
    )


@torch.inference_mode()
def biasbios_instruction_evaluation(
    *,
    mt: models.ModelAndTokenizer,
    dataset: Dataset,
    task: str, 
    prompt_idx: int | list | None = None, 
    pasta_steerer: pasta.PASTA|None = None, 
    emphasized_text: list | None = None, 
    tfidf_vectorizer: TfidfVectorizer | None = None,
    references: dict | None = None,
    batch_size: int = 16,
    top_k: int = DEFAULT_TOP_K,
    max_length: int | None = None,
    max_new_tokens: int | None = None,
    device: Device | None = None,
    desc: str | None = None,
    add_few_shot: int | None = None, 
    few_shot_index: str | None = None,  
    **kwargs,
) -> BiosBiasInstructionEvaluationResults:
    """ Evaluate the instruction following tasks  

    This benchmark evaluation model ability of instruction following 

    Args:
        mt: The model to evaluate. 
        dataset: The dataset to evaluate on.
        task: The name of task (json|pronchange). 
        prompt_idx: The index of prompt template. 
        pasta_steerer: The PASTA steerer to steer the model generaion. 
        emphasized_text: The input spans to be highlighted. 
        tfidf_vectorizer: For computing consistency score.
        references: Mapping from label to reference texts for that label. By default,
            full bios for each label will be used.
        batch_size: Batch size for model.
        top_k: Compute top-k labels predicted by model.
        prompt_key: Which column in dataset to use as prompt.
        max_length: Max sequence length (input+output).
        max_new_tokens: Max number of new tokens to generate. Cannot be used with
            `max_length`, see huggingface docs.
        device: Send model and data to this device.
        desc: TQDM description. 
        add_few_shot: Whether to apply few-shot prompting. 
        few_shot_index: Sample index for few shots. 

    Returns:
        Benchmark results.
    """
    if max_length is None and max_new_tokens is None:
        max_length = DEFAULT_MAX_LENGTH_ERROR_CORRECTION
    if desc is None:
        desc = "Instruction Evaluation"

    if tfidf_vectorizer is None:
        tfidf_vectorizer = data.load_biosbias_tfidf_vectorizer()
    if references is None:
        references = defaultdict(list)
        for sample in dataset:
            references[sample["target_mediated"]].append(sample["source"]["bio"])

    # Load the InstructionEvaluator 
    evaluator = InstructionEvaluator(task, prompt_idx) 

    labels = sorted({x["target_mediated"] for x in dataset})
    label_add_space = True if "gpt" in mt.model.name_or_path else False
    labels_token_idx = precompute.first_token_ids_from_batch(mt, labels, add_space=label_add_space)

    reference_tfidfs = {
        key: tfidf_vectorizer.transform(texts).mean(axis=0).A
        for key, texts in tqdm(references.items(), desc=f"{desc} [reference tfidfs]")
    }

    # If applied, prepare the few-shot demonstration 
    if add_few_shot is not None and add_few_shot > 0: 
        if few_shot_index is not None:
            few_shot_index = [int(idx) for idx in few_shot_index.split(",")]
            assert len(few_shot_index) == add_few_shot
        else:
            few_shot_index = np.random.randint(len(dataset), size=add_few_shot).tolist()
        fewshot_examples = evaluator.prepare_fewshot_examples(dataset.select(few_shot_index))

    columns = data.column_names(dataset, exclude=["target_unmediated"])
    with dataset.formatted_as("torch", columns=columns):
        loader = torch.utils.data.DataLoader(
            cast(torch.utils.data.Dataset, dataset),
            batch_size=batch_size,
        )

        samples = []
        for batch in tqdm(loader, desc=desc):
            ids = batch["id"]
            contexts = batch["context"]
            attributes = batch["attribute"]
            entities = batch["entity"]
            targets = batch["target_mediated"]

            prompts, instructions, entities, (ids, attributes, targets) = evaluator.parapare_prompt_inputs(
                contexts, entities, (ids, attributes, targets)
            )
            targets_idx = precompute.first_token_ids_from_batch(mt, targets, add_space=label_add_space)
            if add_few_shot is not None and add_few_shot > 0:
                prompts = [fewshot_examples+prompt for prompt in prompts]

            with models.set_padding_side(mt, padding_side="left"):
                inputs, prompt_offset_mapping = precompute.inputs_from_batch(mt, prompts, device=device)

            generate_kwargs = dict(
                return_dict_in_generate=True,
                output_scores=True,
                max_length=max_length,
                max_new_tokens=max_new_tokens,
                pad_token_id=mt.tokenizer.eos_token_id,
            )
            if pasta_steerer is not None:
                # Emphasize the selected texts with PASTA 
                if emphasized_text == "instruct":
                    substrings = instructions
                elif emphasized_text == "attribute":
                    substrings = attributes
                else:
                    raise ValueError(f"Unimplemented substrings: {emphasized_text}")
                # Steer the attention scores of selected input spans 
                with pasta_steerer.apply_steering(
                    model=mt.model, 
                    strings=prompts, 
                    substrings=substrings, 
                    model_input=inputs, 
                    offsets_mapping=prompt_offset_mapping
                ) as steered_model: 
                    outputs = steered_model.generate(**inputs, **generate_kwargs)
            else:
                outputs = mt.model.generate(**inputs, **generate_kwargs)

            generations = mt.tokenizer.batch_decode(
                outputs.sequences[:, inputs.input_ids.shape[1] :],
                skip_special_tokens=True,
            )
            distributions = torch.log_softmax(outputs.scores[0], dim=-1)

            for idx, sid, prompt, distribution, generation, target, target_idx in zip(
               range(len(ids)), ids, prompts, distributions, generations, targets, targets_idx
            ):
                label_log_probs = distribution[labels_token_idx]

                logp_predictions, predictions_idx = label_log_probs.topk(
                    k=top_k, dim=-1
                )
                predictions = [labels[idx] for idx in predictions_idx]

                logp_target = distribution[target_idx]

                fluency_score = metrics.weighted_n_gram_entropy(generation)

                [generation_tfidf] = tfidf_vectorizer.transform([generation]).A
                reference_tfidf = reference_tfidfs[target]
                consistency_score = metrics.vector_similarity(
                    generation_tfidf.squeeze(), reference_tfidf.squeeze()
                )

                instruction_evaluation = evaluator.evaluate_sample(generation=generation, target=target)

                sample = InstructionEvaluationSample(
                    id=sid,
                    prompt=prompt,
                    generation=generation,
                    predictions=predictions,
                    logp_predictions=logp_predictions.tolist(),
                    target=target,
                    logp_target=logp_target.item(),
                    fluency=fluency_score,
                    consistency=consistency_score,
                    instruction_evaluation=instruction_evaluation, 
                    sample_attn_scores=None, 
                )
                samples.append(sample)

    n_correct_top1 = sum(x.predictions[0] == x.target for x in samples)
    n_correct_topk = sum(x.target in x.predictions for x in samples)
    top1_accuracy = n_correct_top1 / len(samples)
    topk_accuracy = n_correct_topk / len(samples)

    fluency = metrics.Metric.aggregate([x.fluency for x in samples], store_values=False)
    consistency = metrics.Metric.aggregate(
        [x.consistency for x in samples], store_values=False
    )

    instruction_evaluation = evaluator.aggregate_evaluation_results(samples)
    instruction_evaluation_metrics = InstructionEvaluationMetrics(
        top1_accuracy=top1_accuracy,
        topk_accuracy=topk_accuracy,
        k=top_k,
        fluency=fluency,
        consistency=consistency,
        instruction_evaluation=instruction_evaluation, 
    )

    return BiosBiasInstructionEvaluationResults(
        samples=samples,
        metrics=instruction_evaluation_metrics,
        attentions=None, 
    )



"""Utilities for training models."""
from typing import Sequence, Sized, cast

import torch
from torch.utils import data


class EarlyStopping:
    """Observes a numerical value and determines when it has not improved."""

    def __init__(self, patience: int = 4, decreasing: bool = True):
        """Initialize the early stopping tracker.

        Args:
            patience (int, optional): Allow tracked value to not improve over
                its best value this many times. Defaults to 4.
            decreasing (bool, optional): If True, the tracked value "improves"
                if it decreases. If False, it "improves" if it increases.
                Defaults to True.

        """
        self.patience = patience
        self.decreasing = decreasing
        self.best = float("inf") if decreasing else float("-inf")
        self.num_bad = 0

    def __call__(self, value: float) -> bool:
        """Considers the new tracked value and decides whether to stop.

        Args:
            value (float): The new tracked value.

        Returns:
            bool: True if patience has been exceeded.

        """
        improved = self.decreasing and value < self.best
        improved |= not self.decreasing and value > self.best
        if improved:
            self.best = value
            self.num_bad = 0
        else:
            self.num_bad += 1

        return self.num_bad > self.patience

    @property
    def improved(self) -> bool:
        """Check if the running value just improved."""
        return self.num_bad == 0


TrainValSplit = tuple[data.Subset, data.Subset]


def random_split(dataset: data.Dataset, hold_out: float = 0.1) -> TrainValSplit:
    """Randomly split the dataset into a train and val set.

    Args:
        dataset: The full dataset.
        hold_out: Fraction of data to hold out for the val set. Defaults to .1.

    Returns:
        The train and val sets.

    """
    if hold_out <= 0 or hold_out >= 1:
        raise ValueError(f"hold_out must be in (0, 1), got {hold_out}")

    size = len(cast(Sized, dataset))
    val_size = int(hold_out * size)
    train_size = size - val_size

    for name, size in (("train", train_size), ("val", val_size)):
        if size == 0:
            raise ValueError(
                f"hold_out={hold_out} causes {name} set size " "to be zero"
            )

    splits = data.random_split(dataset, (train_size, val_size))
    assert len(splits) == 2
    train, val = splits
    return train, val


def fixed_split(dataset: data.Dataset, indices: Sequence[int]) -> TrainValSplit:
    """Split dataset on the given indices.

    Args:
        dataset: The dataset to split.
        indices: Indices comprising the right split.

    Returns:
        The subset *not* for the indices, followed by the subset *for* the indices.

    """
    size = len(cast(Sized, dataset))
    for index in indices:
        if index < 0 or index >= size:
            raise IndexError(f"dataset index out of bounds: {index}")

    others = sorted(set(range(size)) - set(indices))
    if not others:
        raise ValueError("indices cover entire dataset; nothing to split!")

    return data.Subset(dataset, others), data.Subset(dataset, indices)


def cosine_similarity_float16(
    x1: torch.Tensor, x2: torch.Tensor, dim: int = -1, eps: float = 1e-5
) -> torch.Tensor:
    """Cosine similarity that is slightly more stable for float16 inputs.

    Prenormalizes each vector, which removes risk of overflow in numerator
    when norms are large.
    """
    x1 = x1 / x1.norm(dim=dim, keepdim=True)
    x2 = x2 / x2.norm(dim=dim, keepdim=True)
    return torch.nn.functional.cosine_similarity(x1, x2, dim=dim, eps=eps)



"""Evaluate editors on the Counterfact benchmark."""
import argparse
import json
import logging
import random
from functools import partial
from pathlib import Path
from typing import cast

from evaluation import benchmarks, data, models, precompute
from evaluation.utils import experiment_utils, logging_utils
from pastalib import pasta

import torch
import torch.utils.data
from tqdm.auto import tqdm

from torch.utils.tensorboard import SummaryWriter

logger = logging.getLogger(__name__)

BENCHMARKS = (
    "efficacy",
    "paraphrase",
    "generation",
)


def _prefix_context(sample: dict) -> dict:
    """Prepend context to all prompts used in the eval."""
    entity = sample["entity"]
    prompt = sample["prompt"]
    context = sample["context"]

    prompt_in_context = precompute.prompt_in_context_from_sample(
        entity, prompt, context
    )

    source = {**sample["source"]}
    for key in ("generation_prompts", "paraphrase_prompts"):
        source[key] = [
            precompute.prompt_in_context_from_sample(entity, other_prompt, context)
            for other_prompt in source[key]
        ]
    return {"source": source, "prompt": prompt_in_context}


def main(args: argparse.Namespace) -> None:
    """Run the CounterFact benchmark."""
    experiment = experiment_utils.setup_experiment(args)
    logging_utils.configure(args=args)
    data.disable_caching()
    if args.debug: 
        experiment_utils.set_ipdb_trace()

    # Initialize the model and tokenizer 
    device = args.device or "cuda" if torch.cuda.is_available() else "cpu"
    mt = models.load_model(args.model, device=device, fp16=args.fp16)
    # Set up the PASTA steerer 
    if args.apply_pasta:
        head_config = experiment_utils.read_head_config(args.pasta_head_config)
        pasta_steerer = pasta.PASTA(
            mt.model, 
            mt.tokenizer,
            head_config=head_config, 
            alpha=args.alpha, 
            scale_position=args.scale_position,
        )
    else:
        pasta_steerer = None

    # Set up the evaluation data 
    logger.info("loading several data sources")
    if args.example_subset is not None:
        split = f"train[{args.example_subset}]"
    else:
        split = "train[5000:10000]"
    dataset = data.load_dataset("counterfact", split=split)
    dataset = precompute.from_args(args, dataset)
    attribute_snippets = data.load_attribute_snippets()
    tfidf_vectorizer = data.load_counterfact_tfidf_vectorizer()

    dataset = precompute.editor_inputs_from_dataset(
            dataset=dataset,
            mt=mt,
            return_entity_hiddens=False,
            return_attribute_hiddens=False,
            return_token_ranges=False,
            return_target_token_ids=True,
            target_token_first_space=True if "gptj" in args.model else False, 
            desc="precompute target token ids",
    )
    dataset = dataset.map(_prefix_context, desc="prefix context")

    # Set up the benchmark arguments 
    benchmark_kwargs: dict = dict(dataset=dataset, device=device)
    benchmark_kwargs["mt"] = mt
    benchmark_kwargs["pasta_steerer"] = pasta_steerer 
    output_dir = "{pasta}".format(
        pasta = f"pasta-{args.alpha}-{args.scale_position}" if args.apply_pasta else "baseline",
    )
    if args.add_few_shot:
        output_dir = output_dir + "_few-%d-%s"%(args.add_few_shot, str(args.few_shot_index))
    if args.add_marker:
        output_dir = output_dir + "_add-marker-%s"%(args.add_marker)
    results_output_dir = (
        experiment.results_dir 
        / f"{args.model.split('/')[-1]}"
        / output_dir 
    )
    if args.pasta_head_config is not None:
        results_output_dir = results_output_dir / f"{args.pasta_head_config.split('/')[-1]}"
    results_output_dir.mkdir(exist_ok=True, parents=True)
    tb_writter = SummaryWriter(log_dir=str(results_output_dir.resolve()))

    results: (
        benchmarks.EfficacyBenchmarkResults
        | benchmarks.CounterFactParaphraseBenchmarkResults
        | benchmarks.CounterFactGenerationBenchmarkResults
        | benchmarks.EssenceBenchmarkResults
    )
    logger.info(f"eval counterfact")
    
    for benchmark_name in args.benchmarks:
        results_file = results_output_dir / f"{benchmark_name}.json"
        if results_file.exists() and not args.overwrite_output_dir:
            logger.info(
                f"found existing {benchmark_name} results "
                f"at {results_file}"
            )
            continue

        benchmark_kwargs["add_unmediated_fact"] = args.add_unmediated_fact 
        benchmark_kwargs["batch_size"] = args.batch_size
        benchmark_kwargs["max_length"] = args.max_length 
        benchmark_kwargs["add_few_shot"] = args.add_few_shot 
        benchmark_kwargs["few_shot_index"] = args.few_shot_index 
        benchmark_kwargs["add_marker"] = args.add_marker 

        if benchmark_name == "efficacy":
            results = benchmarks.counterfact_efficacy(**benchmark_kwargs)
        elif benchmark_name == "paraphrase":
            results = benchmarks.counterfact_paraphrase(**benchmark_kwargs)
        elif benchmark_name == "generation":
            results = benchmarks.counterfact_generation(
                attribute_snippets=attribute_snippets,
                tfidf_vectorizer=tfidf_vectorizer,
                **benchmark_kwargs,
            )
        else:
            raise ValueError(f"unknown benchmark: {benchmark_name}")

        logging.info(
            f"{benchmark_name} benchmark complete! results:\n%s",
            json.dumps(results.metrics.to_dict(), indent=1),
        )
        for key, value in results.metrics.to_dict().items():
            tb_writter.add_scalar(f"{benchmark_name}/{key}", value['mean'], 1)
        
        with results_file.open("w") as handle:
            json.dump(results.to_dict(), handle)

        metrics_file = results_file.parent / f"{benchmark_name}_metrics.json"
        with metrics_file.open("w") as handle:
            json.dump(results.metrics.to_dict(), handle)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="evaluate editors")
    parser.add_argument(
        "--benchmarks",
        "-b",
        nargs="+",
        choices=BENCHMARKS,
        default=BENCHMARKS,
        help="benchmarks to run, defaults depend on dataset",
    )
    models.add_model_args(parser)
    precompute.add_preprocessing_args(parser)
    experiment_utils.add_experiment_args(parser)
    logging_utils.add_logging_args(parser)
    args = parser.parse_args()
    main(args)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = [
    "transformers>=4.31.1"
]

setup(
    name="pastalib",
    version="0.1.3",
    author="Qingru Zhang, Chandan Singh, Lucas Liu, Xiaodong Liu, Bin Yu, Jianfeng Gao, Tuo Zhao",
    author_email="qingru.zhang@gatech.edu",
    description="PyTorch Implementation for PASTA, A Post-hoc Attention Steering Approach that enables users to emphasize specific contexts for LLMs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/QingruZhang/PASTA",
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.7.0',
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)


"""Functions for loading and interacting with pretrained language models.

This module is designed to house all the annoying branching logic
that comes with supporting analysis of many slightly different model
implementations.
"""
import argparse
import logging
import json 
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Iterator, Literal, Optional, Sequence, overload

from evaluation.utils import tokenizer_utils
from evaluation.utils.typing import Device, Model, Tokenizer

import torch
import transformers

logger = logging.getLogger(__name__)

GPT_J_NAME_SHORT = "gptj"  # A useful alias for the CLI.
GPT_J_NAME = "EleutherAI/gpt-j-6B"

GPT_NEO_X_NAME_SHORT = "neox"
GPT_NEO_X_NAME = "EleutherAI/gpt-neox-20b"


@dataclass(frozen=True)
class ModelAndTokenizer:
    """A pretrained model and its tokenizer."""

    model: Model
    tokenizer: Tokenizer

    def to_(self, device: Optional[Device]) -> None:
        """Send model to the device."""
        self.model.to(device)

    def eval_(self) -> None:
        """Set model to eval mode."""
        self.model.eval()


def unwrap_model(value: Model | ModelAndTokenizer) -> Model:
    """Unwrap the model if necessary."""
    if isinstance(value, ModelAndTokenizer):
        return value.model
    return value


def unwrap_tokenizer(tokenizer: ModelAndTokenizer | Tokenizer) -> Tokenizer:
    """Unwrap the tokenizer."""
    if isinstance(tokenizer, ModelAndTokenizer):
        return tokenizer.tokenizer
    return tokenizer


def determine_layers(model: ModelAndTokenizer | Model) -> tuple[int, ...]:
    """Return all hidden layer names for the given model."""
    model = unwrap_model(model)
    assert isinstance(model, Model)

    if isinstance(model, transformers.GPTNeoXForCausalLM):
        n_layer = model.config.num_hidden_layers
    else:
        n_layer = model.config.n_layer

    return (*range(n_layer),)


@overload
def determine_layer_paths(
    model: ModelAndTokenizer | Model,
    layers: Optional[Sequence[int]] = ...,
    *,
    return_dict: Literal[False] = ...,
) -> Sequence[str]:
    """Determine layer path for each layer."""
    ...


@overload
def determine_layer_paths(
    model: ModelAndTokenizer | Model,
    layers: Optional[Sequence[int]] = ...,
    *,
    return_dict: Literal[True],
) -> dict[int, str]:
    """Determine mapping from layer to layer path."""
    ...


def determine_layer_paths(
    model: ModelAndTokenizer | Model,
    layers: Optional[Sequence[int]] = None,
    *,
    return_dict: bool = False,
) -> Sequence[str] | dict[int, str]:
    """Determine the absolute paths to the given layers in the model.

    Args:
        model: The model.
        layers: The specific layer (numbers) to look at. Defaults to all of them.
        return_dict: If True, return mapping from layer to layer path,
            otherwise just return list of layer paths in same order as `layers`.

    Returns:
        Mapping from layer number to layer path.

    """
    model = unwrap_model(model)

    if layers is None:
        layers = determine_layers(model)

    assert isinstance(model, Model), type(model)

    layer_paths = {}
    for layer in layers:
        if isinstance(model, transformers.GPTNeoXForCausalLM):
            layer_path = f"gpt_neox.layers.{layer}"
        else:
            layer_path = f"transformer.h.{layer}"
        layer_paths[layer] = layer_path

    return layer_paths if return_dict else tuple(layer_paths[la] for la in layers)


def determine_hidden_size(model: ModelAndTokenizer | Model) -> int:
    """Determine hidden rep size for the model."""
    model = unwrap_model(model)
    return model.config.hidden_size


def determine_device(model: ModelAndTokenizer | Model) -> torch.device | None:
    """Determine device model is running on."""
    parameter = any_parameter(model)
    return parameter.device if parameter is not None else None


def determine_dtype(model: ModelAndTokenizer | Model) -> torch.dtype | None:
    """Determine dtype of model."""
    parameter = any_parameter(model)
    return parameter.dtype if parameter is not None else None


def any_parameter(model: ModelAndTokenizer | Model) -> torch.nn.Parameter | None:
    """Get any example parameter for the model."""
    model = unwrap_model(model)
    return next(iter(model.parameters()), None)


@contextmanager
def set_padding_side(
    tokenizer: Tokenizer | ModelAndTokenizer, padding_side: str = "right"
) -> Iterator[None]:
    """Wrap `tokenizer_utils.set_padding_side`."""
    tokenizer = unwrap_tokenizer(tokenizer)
    with tokenizer_utils.set_padding_side(tokenizer, padding_side=padding_side):
        yield


def map_to(
    orig: Any, device: Device | None = None, dtype: torch.dtype | None = None
) -> Any:
    """Map all tensors in the given value to the device.

    Args:
        orig: Any sequence of or mapping to tensors, or just a tensor.
        device: Device to send to.

    Returns:
        Same value, but with all tensors moved to the device.

    """
    if device is None and dtype is None:
        return orig

    result = orig
    if isinstance(orig, torch.Tensor):
        result = orig.to(device=device, dtype=dtype)
    elif isinstance(orig, dict):
        result = {
            key: map_to(value, device=device, dtype=dtype)
            for key, value in orig.items()
        }
    elif isinstance(orig, (list, tuple)):
        result = orig.__class__(
            map_to(value, device=device, dtype=dtype) for value in orig
        )
    assert isinstance(result, orig.__class__), f"{type(result)}/{type(orig)}"
    return result


def load_model(
    name: str, 
    device: Optional[Device] = None, 
    fp16: Optional[bool] = None, 
) -> ModelAndTokenizer:
    """Load the model given its string name.

    Args:
        name: Name of the model.
        device: If set, send model to this device. Defaults to CPU.
        fp16: Whether to use half precision. If not set, depends on model.

    Returns:
        ModelAndTokenizer: Loaded model and its tokenizer.

    """
    if name == GPT_J_NAME_SHORT:
        name = GPT_J_NAME
    elif name == GPT_NEO_X_NAME_SHORT:
        name = GPT_NEO_X_NAME

    is_gpt_j_variant = name == GPT_J_NAME or GPT_J_NAME_SHORT in name
    is_neo_x_variant = name == GPT_NEO_X_NAME or GPT_NEO_X_NAME_SHORT in name
    if fp16 is None:
        fp16 = is_gpt_j_variant or is_neo_x_variant
    torch_dtype = torch.float16 if fp16 else None
    model_kwargs: dict = dict(torch_dtype=torch_dtype)
    if name == GPT_J_NAME or GPT_J_NAME_SHORT in name:
        model_kwargs["low_cpu_mem_usage"] = True
        if fp16:
            model_kwargs["revision"] = "float16"
    if "llama" in name:
        model_kwargs["torch_dtype"] = torch.bfloat16 
    if device=="auto":
        model_kwargs["device_map"] = "auto"
        model_kwargs["torch_dtype"] = torch.bfloat16 

    logger.info(f"loading {name} (device={device}, fp16={fp16})")

    config = transformers.AutoConfig.from_pretrained(name)
    model = transformers.AutoModelForCausalLM.from_pretrained(
        name, config=config, **model_kwargs
    )

    if is_neo_x_variant:
        model.to(torch_dtype)
    if device != "auto":
        model.to(device).eval()

    tokenizer = transformers.AutoTokenizer.from_pretrained(name)
    tokenizer.pad_token = tokenizer.eos_token

    return ModelAndTokenizer(model, tokenizer)


def add_model_args(parser: argparse.ArgumentParser) -> None:
    """Add args needed to load a model.

    The args include:
        --model: The language model to load, defaulting to GPT-J.
        --device: The device to send model and inputs to.
        --fp16: Whether to use half precision version of the model.
            Note this is used as `--fp16 False` since default value depends on
            which model we are loading.
    """
    parser.add_argument(
        "--model",
        "-m",
        default=GPT_J_NAME_SHORT,
        help="model to edit",
    )

    parser.add_argument("--device", help="device to train on")
    parser.add_argument("--fp16", type=bool, help="set whether to use fp16")
